#include <MIDI.h>               // MIDI Library 4.1
#include <midi_Defs.h>
#include <midi_Namespace.h>
#include <midi_Settings.h>

#include <avr/interrupts.h>
#include <pgmspace.h>           // For any tables in ROM
#include <SPI.h>                // Used to communicate with the DAC
#include <DAC_MCP49xx.h>        // MCP-series SPI DAC library

// Libraries for the Adafruit RGB/LCD Shield
#include <Wire.h>
#include <Adafruit_MCP23017.h>
#include <Adafruit_RGBLCDShield.h>

/* Old Crow's Crowminius Synthesizer
 
   Crowminius v1 by Scott Rider, aka The Old Crow
   cc-by-sa ** OPEN SOURCE: FREE FOR NON-COMMERCIAL USE **
   
   Acknowledgments:
   Jurgen Haible (RIP) -- I will always pay it forward.
   Trevor Page -- the RX303a PIC code for your MIDI to CV/gate is awesome!
   Martin Nawrath -- Arduino DDS sine wave generator example code
   Various Arduino and C folks -- thanks!   

   Revision: 2.1
   LAST MODIFIED 02-JUN-2015 20:19
   
   Receives MIDI currently on MIDI channel 1. Each key found will load the DAC
   with note data for the most recently pressed key and enable the GATE signal.
   As keys are released, MIDI notes will play legato-style (no re-trigger). The
   GATE is only turned off once all MIDI keys are released.

   The circuit uses a dual 12-bit SPI DAC that has an internal 2.048v voltage
   reference. The DAC can also be set to use a vref gain of 2, so a 4.096v full-scale
   output is available. The first DAC channel provides a voltage for the pitch
   control and is scaled to 0.815V/Octave to allow MIDI notes over a range
   of five octaves: 0v is F1, 4.075v is F6 (61 notes total).
   
   The second DAC channel is used to allow a filter modulation CV to appear based on
   selected MIDI sources such as key velocity, aftertouch, etc.
   
   Note: V-USB not implemented yet for MIDI over USB.
   
*/

#define  bGrab  pgm_read_byte_near // Define shorter names for these progmem calls
#define  wGrab  pgm_read_word_near //  "   "

#define DACCS 10 // Arduino pin 10 / 328P pin 16 DACCS (chip select)
#define LDACS 7  // Arduino pin 7 / 328P pin 13 LDAC (latch enable)

// Tell DAC library which part and pins to use
// Actually using an MCP4822 which is the 4922 with internal 2.048v Vref.
DAC_MCP49xx dac(DAC_MCP49xx::MCP4922, DACCS, LDACS); // DAC initialize
/*
  dac.setBuffer(bool) true to enable external Vref buffers, off by default --NOT ON 4822
  dac.setGain(int) set DAC gain 1x or 2x, use value of 1 (default) or 2
  dac.output2(ushort val1, ushort val2) to send values to dac, calls dac.latch by default
  dac.outputA(ushort) to set output A (must use dac.latch to set voltage on pin)
  dac.outputB(ushort) to set output B (must use dac.latch to set votlage on pin)
  dac.latch() to set DAC outputs
*/

// For MIDI 4.x backward compatibility -- uncomment next line for MIDI library 4.2 or higher.
// MIDI_CREATE_DEFAULT_INSTANCE();

// Tell Adafruit LCD library an LCD might be connected
Adafruit_RGBLCDShield lcd = Adafruit_RGBLCDShield();

//
// Pin Definitions
const int usbint = 2;  // 328P pin 4,  USB D+ interrupt for V-USB
const int a440pw = 3;  // 328P pin 5,  PWM output for A440 sine tone
const int dipsw1 = 4;  // 328P pin 6   Option DIP switch #1
const int dipsw2 = 5;  // 328P pin 11  Option DIp switch #2
const int a440sw = 6;  // 328P pin 12, A440 on/off switch input
const int gate   = 8;  // 328P pin 14, synth gate out
const int pwmmod = 9;  // 328P pin 15, MIDI modwheel PWMDAC out **USES TIMER1**
const int usbneg = 14; // 328P pin 23, USB D- data pin for V-USB
const int usbpos = 15; // 328P pin 24, USB D+ data pin for V-USB
const int usbpul = 16; // 328P pin 25, USB pullup pin for V-USB

// Lookup tables in program flash area
// Table of 256 sine values stored as one sine period.
PROGMEM  prog_uchar sineTable[]  = {
  127,130,133,136,139,143,146,149,152,155,158,161,164,167,170,173,176,178,181,184,187,190,192,195,
  198,200,203,205,208,210,212,215,217,219,221,223,225,227,229,231,233,234,236,238,239,240,242,243,
  244,245,247,248,249,249,250,251,252,252,253,253,253,254,254,254,254,254,254,254,253,253,253,252,
  252,251,250,249,249,248,247,245,244,243,242,240,239,238,236,234,233,231,229,227,225,223,221,219,
  217,215,212,210,208,205,203,200,198,195,192,190,187,184,181,178,176,173,170,167,164,161,158,155,
  152,149,146,143,139,136,133,130,127,124,121,118,115,111,108,105,102,99,96,93,90,87,84,81,78,76,73,
  70,67,64,62,59,56,54,51,49,46,44,42,39,37,35,33,31,29,27,25,23,21,20,18,16,15,14,12,11,10,9,7,6,5,
  5,4,3,2,2,1,1,1,0,0,0,0,0,0,0,1,1,1,2,2,3,4,5,5,6,7,9,10,11,12,14,15,16,18,20,21,23,25,27,29,31,
  33,35,37,39,42,44,46,49,51,54,56,59,62,64,67,70,73,76,78,81,84,87,90,93,96,99,102,105,108,111,115,
  118,121,124
};

// Variables

// LOWNOTE selects which note is to represent 0v on the DAC. Note #29 is F1.
const uint16_t LOWNOTE = 29; // Transposes CVs down 29 semitones ~= 2.5 octaves (note F1)
const uint16_t HIGHNOTE = 89; // Highest note is F6
// Scale factor of 273 (/4) is for 0.815V/Oct keyboard scale over the 0v to 4.096v DAC range.(5 oct)
// Scale factor of 341 (/4) is for 1.00v/Oct keyboard scale over the 0v to 4.096v DAC range. (4 oct, use HIGHNOTE=77)
// Example, for a 5-octave keyboard, there are 60 notes pluse the 1st note at zero volts.
// To scale the DAC the scale factor would be 4096/60*4 = 273.  The x4 is to allow
// for decimation of the data in the bit-precision statement of outputCV() below.
const uint16_t SCALEFACTOR = 273; // Scale factor depends on DAC full scale resolution

byte channel = 1; // MIDI channel, default is 1.
byte midiNoteNum;
byte midiNoteVel;
byte midiCtrlrCode;
byte midiModWhl;
byte midiNoteLegato;

boolean a440Switch, a440SwitchLastState = false; // For reading the A440 physical switch
boolean a440Latch, a440LatchLastState = false; // Build a latch for the A440 state machine
long a440LastDebounce = 0;
long debounceDelay = 50; // 50ms

// DDS sine variables
double dfreq;  // Set up dfreq to be adjustable w/default of 440.0Hz (assigned below)
// const double refclk=31372.549; // =16MHz / 510
const double refclk = 31376.6;      // measured
// Variables used inside DDS interrupt service routine declared as volatile.
volatile byte indexOffset;        // Offset into sine table
volatile unsigned long phaseAcc;  // phase accumulator
volatile unsigned long tuneWord;  // dds tuning word


// Buffer management variables are preloaded as "clear" (no notes to play)
int midiNoteBufferIndex = 0;
byte midiNoteBuffer[] = { 128, 128, 128, 128, 128, 128, 128, 128, 128, 128 }; // Padded 8-note buffer for MIDI
byte midiNoteToRemove = 255;

// Flush the midiNoteBuffer, used by MIDI controller code 123 (all notes off) as
// well as initialization. Gate is set to inactive.
void clearMidiNoteBuffer() {
        for (int i = 0; i < 10; i++)
                midiNoteBuffer[i] = 128;
        midiNoteBufferIndex = 0;
        midiNoteToRemove = 255;
        digitalWrite(gate, LOW);
}

// Called at the end of processing MIDI note off messages to repack the midiNoteBuffer
void compactMidiNoteBuffer() {
        for (int i = 1; i < 9; i++) {
            if (midiNoteBuffer[i] == midiNoteToRemove) { // Maintain midiNoteBuffer as needed
                midiNoteBuffer[i] = 128; // Clear note entry
                midiNoteToRemove = 255; // Unflag the now-removed note
                midiNoteBufferIndex--; // Shrink filled buffer count
            }
            if (midiNoteBuffer[i] == 128) { // If current buffer location empty
                midiNoteBuffer[i] = midiNoteBuffer[i + 1]; // compact buffer
                midiNoteBuffer[i + 1] = 128; // justify with "slot empty" value
            }
        }
}

// Read the two DIP switches
byte readDIPSwitches() {
      byte switches = 0;
      switches |= digitalRead(dipsw1); // Read 1st switch
      switches |= digitalRead(dipsw2) << 1; // Read 2nd switch
      return switches;
}

void outputCV(byte note) {
      // V/Oct DAC mode, Minimoog had a 44-note keyboard--60 notes scale here **starting at F1**
      while (note > HIGHNOTE) note -= 12;  // Wrap notes above F6 to highest usable octave
      while (note < LOWNOTE) note += 12;  // Wrap notes below F1 to lowest usable octave
      uint16_t x = (note - LOWNOTE) * SCALEFACTOR; // Magic for correct semitone scale, ((note-LOWNOTE)*(4096*4)/60)
      uint16_t noteOut =  (x >> 2) + ((x >> 1) & 1);  // Round to nearest int for 0.5 LSB accuracy
      dac.outputA(noteOut); // Exponential CV uses scaled note code with note DAC Vref=4.096v
      dac.latch(); // Set CV on DAC pin
}

void disableA440() {
      lcd.print(F("Crowminius v1.1"));
      TIMSK2 &= ~(1 << TOIE2); // Disable timer 2 interrupt (sine)
}

void enableA440() {
      lcd.print(F("Crowminius A440"));
      TIMSK2 |= (1 << TOIE2);  // Enable timer 2 interrupt (sine)
}
  
// Timer2 interrupt service every 1/31372.550 KHz = 32uSec
// This is the timebase reference for the DDS generator
// Freq = (TUNEWORD * REFCLK) / (2 exp 32)

ISR(TIMER2_OVF_vect) {
  phaseAcc = phaseAcc + tuneWord;      // soft DDS, phase accumulator with 32 bits
  indexOffset = phaseAcc >> 24;       // use upper 8 bits of phase accumulator as waveform index
  OCR2B = bGrab(sineTable + indexOffset); // grab value from sine table and send to PWM DAC
}

void setup() {  	

        pinMode(gate, OUTPUT);
        pinMode(a440pw, OUTPUT);
        pinMode(pwmmod, OUTPUT);
        pinMode(a440sw, INPUT);
        pinMode(dipsw1, INPUT);
        pinMode(dipsw2, INPUT);
        
// Initialize note buffer
        clearMidiNoteBuffer();

// Timer2 setup
// Set prescaler to 1, PWM mode to phase correct PWM,  16000000/510 = 31372.55 Hz clock

        noInterrupts();
        TCCR2A = 0;
        TCCR2B = 0;
        TCNT2 = 0;
        TCCR2B |= (1 << CS20);    // Set timer2 clock prescaler to 1
        TCCR2A |= (1 << COM2B1);  // /
        TCCR2A |= (1 << WGM20);   // Mode 1  / Phase Correct PWM
        TIMSK2 &= ~(1 << TOIE2);  // Disable timer 2 interrupt by default.
        interrupts();

        dfreq=440.0;              // initial output frequency = 440.0 Hz
        tuneWord = pow(2,32) * dfreq / refclk; // calulate DDS tuning word 

// Initialize MIDI and DAC stuff.

        MIDI.begin(channel); // MIDI receive on default channel 1
        MIDI.turnThruOn(); // Use thru transmit for active LED MIDI status
        dac.setGain(2);   // DAC gain x2 for full scale = 4.096v
        dac.outputB(0);  // Set filter mod DAC to zero/off.
        dac.latch();  // DAC outputs set on latch()
        analogWrite(pwmmod, 0); // MIDI Mod Wheel "off"

// Initialize LCD DiSplay 

        lcd.begin(16, 2);
        lcd.setBacklight(1);
        lcd.print(F("Crowminius v1.1"));
        lcd.setCursor(0, 1);
        lcd.print(F("MIDI: 90xxxx"));

//	delay(100); // Short delay for power-up of analog circuits
}

void loop() {
        
        // Process MIDI notes, if any (each MIDI.read is one 2 or 3-byte message)
        if (MIDI.read()) {
            switch(MIDI.getType())
            {
                case midi::NoteOn:
                        midiNoteNum = MIDI.getData1();
                        midiNoteVel = MIDI.getData2();
                        lcd.setCursor(8, 1);
                        lcd.print(midiNoteNum, HEX); // Show most recent MIDI note, velocity
                        lcd.print(midiNoteVel, HEX);
                        if (midiNoteVel != 0) {
                                midiNoteBufferIndex++;
                                midiNoteBuffer[midiNoteBufferIndex] = midiNoteNum;  // Add newest note to buffer
                                if ((digitalRead(dipsw1) == 0) && (digitalRead(dipsw2) == 0)) // DIP switch 1 on is CC/vel
                                        dac.outputB(midiNoteVel << 5); // DIP switch 2 selects velocity->VCF or expression pedal->VCF
                                outputCV(midiNoteNum); // Send to DAC and latch
                                digitalWrite(gate, HIGH); // Key on
                                if (midiNoteBufferIndex == 9) { // 9 notes in buffer?  Limit it to 8
                                        midiNoteBuffer[1] = 128; // Flag oldest entry for removal
                                        midiNoteBufferIndex--; // Back out to 8th entry
                                }
                        } else { // "Velocity zero" note off
                                midiNoteToRemove = midiNoteNum; // Flag note to clear from buffer
                                if (midiNoteBuffer[midiNoteBufferIndex] == midiNoteToRemove) { // If current note is to be turned off
                                        if (midiNoteBufferIndex > 0) { // then provided the buffer is not empty
                                                if ((midiNoteBufferIndex - 1) > 0) { // and it is not the only note playing
                                                        midiNoteLegato = midiNoteBuffer[midiNoteBufferIndex - 1]; // play the previous note
                                                        outputCV(midiNoteLegato); // legato style (no re-trigger)
                                                }
                                        }
                                }
                                compactMidiNoteBuffer(); // Manage MIDI note buffer
                        }
                break;
                case midi::NoteOff: // Same action as velocity zero note off
                    midiNoteToRemove = MIDI.getData1(); // Flag note to clear from buffer
                    if (midiNoteBuffer[midiNoteBufferIndex] == midiNoteToRemove) { // If current note is to be turned off
                        if (midiNoteBufferIndex > 0) { // then provided the buffer is not empty
                            if ((midiNoteBufferIndex - 1) > 0) { // and it is not the only note playing
                                midiNoteLegato = midiNoteBuffer[midiNoteBufferIndex - 1]; // play the previous note legato style (no re-trigger)
                                outputCV(midiNoteLegato);
                            }
                        }
                    }
                    compactMidiNoteBuffer(); // Manage MIDI note buffer
                break;
                case midi::ControlChange:
                        midiCtrlrCode = MIDI.getData1();
                        if (midiCtrlrCode == 123 && MIDI.getData2() == 0) // CC#123 with data2=0 is "all notes off"
                                clearMidiNoteBuffer;
                        if (midiCtrlrCode == 1) { // Mod Wheel CC
                                midiModWhl = MIDI.getData2();
                                midiModWhl = midiModWhl << 1; // Mod Wheel value is 7 bits
                                analogWrite(pwmmod, midiModWhl);
                        }
                        // DIP switch #1 on (low) is CC/Vel, DIP switch #2 off (high) is CC#11, expression pedal.
                        if ((digitalRead(dipsw1) == 0) && (digitalRead(dipsw2) != 0) && (midiCtrlrCode == 11)) {
                                uint16_t midiFCV = MIDI.getData2() << 5; // FCV value is 7 bits, shift to upper 8 bits of 12-bit DAC, keep a copy.
                                dac.outputB(midiFCV); // Write filter mod data to 2nd DAC channel
                                dac.latch(); // Latch a continous controller msg on the DAC.
                        } 
                break;
                case midi::AfterTouchChannel:
                        if (digitalRead(dipsw1) != 0) { // DIP switch 1 off (high) is channel pressure cc (aftertouch)
                                dac.outputB(MIDI.getData1() << 5); // Move 0~127 into 12-bit range for 2nd DAC channel. No copy needed.
                                dac.latch(); // Latch a continous controller msg on the DAC.
                        }
                break;
                default:
                break;
            }
        } // End of MIDI input processing

        // A440 switch needs a toggled latch bit so as to enable/disable the reference tone once per switch change.
        a440Switch = digitalRead(a440sw);
        if (a440Switch != a440SwitchLastState) { // If state changed, switch was moved
                a440LastDebounce = millis(); // Grab a debounce reference count
                a440SwitchLastState = a440Switch; // Remember the new switch state
        }
        if ((millis() - a440LastDebounce) > debounceDelay) { // 50ms settling time for switch change
                if (a440Switch != a440LatchLastState) { 
                        a440Latch = !a440Latch; // latch bit toggled
                        lcd.setCursor(0, 0);
                        a440Latch ? enableA440(): disableA440();
                }
                a440LatchLastState = a440Switch; // Update enable/disable latch bit
        }
        // Check midiNoteBufferIndex and if zero (no keys down), release the gate.
        if (midiNoteBufferIndex == 0)
                digitalWrite(gate, LOW);  //Release GATE
}                      
// END

